import bs
import bsInternal
from bsVector import Vector
import weakref
import random
import bsUtils

class Team(object):
    """
    category: Game Flow Classes
    
    A team of one or more bs.Players.
    Note that a player *always* has a team;
    in some cases, such as free-for-all bs.Sessions,
    the teams consists of just one bs.Player each.

    Attributes:
    
       name
          The team's name.

       color
          The team's color.

       players
          The list of bs.Players on the team.

       gameData
          A dict for use by the current bs.Activity
          for storing data associated with this team.
          This gets cleared for each new bs.Activity.

       sessionData
          A dict for use by the current bs.Session for
          storing data associated with this team.
          Unlike gameData, this perists for the duration
          of the session.
    """
    
    def __init__(self,teamID=0,name='',color=(1,1,1)):
        """
        Instantiate a team. In most cases teams are provided to you
        automatically by the bs.Session, so calling this shouldn't be necessary.
        """
        # we override __setattr__ to lock ourself down
        # so we have to set attrs all funky-like
        object.__setattr__(self,'_teamID',teamID)
        object.__setattr__(self,'name',name)
        object.__setattr__(self,'color',color)
        object.__setattr__(self,'players',[])
        object.__setattr__(self,'gameData',{}) # per-game user-data
        object.__setattr__(self,'sessionData',{}) # per-session user-data
    def getID(self):
        'Returns the numeric team ID.'
        return object.__getattribute__(self,'_teamID')
    def _reset(self):
        object.__setattr__(self,'players',[])
        object.__setattr__(self,'gameData',{})
    def __setattr__(self,name,value):
        raise Exception("can't set attrs on bs.Team objects")

class OutOfBoundsMessage(object):
    """
    category: Message Classes

    Tells an object that it is out of bounds.
    """
    pass

class DieMessage(object):
    """
    category: Message Classes

    Tells an object to die.
    Most bs.Actors respond to this.

    Attributes:

       immediate
          Whether the target must disappear immediately or can take its time
          (leaving a corpse, transitioning out, etc).

       how
          The particular reason for death; 'fall', 'impact', 'leftGame', etc.
          This can be examined for scoring or other purposes.
         
    """
    def __init__(self,immediate=False,how="generic"):
        """
        Instantiate with the given values.
        """
        self.immediate = immediate
        self.how = how

class StandMessage(object):
    """
    category: Message Classes

    Tells an object to position itself to be standing upright at the given
    position. Used when teleporting players to home base, etc.

    Attributes:

       position
          Where to stand.

       angle
          The angle to face (in degrees)
    """
    def __init__(self,position=(0,0,0),angle=0):
        """
        Instantiate with a given position and angle.
        """
        self.position = position
        self.angle = angle

class PickUpMessage(object):
    """
    category: Message Classes
    
    Tells an object that it has picked something up.

    Attributes:

       node
          The bs.Node that is getting picked up.
    """
    def __init__(self,node):
        'Instantiate with a given bs.Node.'
        self.node = node

class DropMessage(object):
    """
    category: Message Classes
    
    Tells an object that it has dropped whatever it was holding.
    """
    pass

class PickedUpMessage(object):
    """
    category: Message Classes
    
    Tells an object that it has been picked up by something.

    Attributes:
    
       node
          The bs.Node doing the picking up.
    """
    def __init__(self,node):
        """
        Instantiate with a given bs.Node.
        """
        self.node = node

class DroppedMessage(object):
    """
    category: Message Classes
    
    Tells an object that it has been dropped.

    Attributes:
    
       node
          The bs.Node doing the dropping.
    """
    def __init__(self,node):
        """
        Instantiate with a given bs.Node.
        """
        self.node = node

class ShouldShatterMessage(object):
    """
    category: Message Classes
    
    Tells an object that it should shatter.
    """
    pass

class ImpactDamageMessage(object):
    """
    category: Message Classes
    
    Tells an object that it has been jarred violently and
    may want to be damaged.

    Attributes:

        intensity
            The intensity of the impact.
    """
    def __init__(self,intensity):
        """
        Instantiate a messages with a given intensity value.
        """
        self.intensity = intensity

class FreezeMessage(object):
    """
    category: Message Classes
    
    Tells an object to become frozen
    (as in the effects of an ice bs.Bomb).
    """
    pass

class ThawMessage(object):
    """
    category: Message Classes
    
    Tells an object that was frozen by a bs.FrozenMessage
    to thaw out.
    """
    pass

class HitMessage(object):
    """
    category: Message Classes
    
    Tells an object it has been hit in some way.
    This is used by punches, explosions, etc to convey
    their effect to a target.
    """
    def __init__(self,srcNode=None,pos=Vector(0,0,0),velocity=Vector(0,0,0),
                 magnitude=1.0,velocityMagnitude=0.0,radius=1.0,sourcePlayer=None,
                 kickBack=1.0,flatDamage=None,hitType='generic',forceDirection=None,
                 hitSubType='default'):
        """
        Instantiate a message with various bits of information
        on the type of hit that occurred.
        """
        self.srcNode = srcNode
        self.pos = pos
        self.velocity = velocity
        self.magnitude = magnitude
        self.velocityMagnitude = velocityMagnitude
        self.radius = radius
        self.sourcePlayer = sourcePlayer
        self.kickBack = kickBack
        self.flatDamage=flatDamage
        self.hitType=hitType
        self.hitSubType=hitSubType
        self.forceDirection=forceDirection if forceDirection is not None else velocity

class Actor(object):
    """
    category: Game Flow Classes

    Actors are high level organizational entities that generally manage one or
    more bs.Nodes, bits of game media, etc, along with the associated logic to
    wrangle them.

    If bs.Nodes represent cells, think of bs.Actors as organs.
    Some example actors include bs.Bomb, bs.Flag, and bs.Spaz.

    One key feature of actors is that they generally 'die'
    (killing off or transitioning out their nodes) when the last python
    reference to them disappears, so you can use logic such as:

    # create a flag in our activity
    self.flag = bs.Flag(translate=(0,10,0))

    # later, destroy the flag..
    # (provided nothing else is referencing it)
    self.flag = None

    This is in contrast to the behavior of the more low level bs.Nodes,
    which are always explicitly created and destroyed regardless of how many python
    references to them exist.

    Another key feature of bs.Actor is its handleMessage() method, which
    takes a single arbitrary object as an argument. This provides a safe way
    to communicate between bs.Actor, bs.Activity, bs.Session, and any other
    object providing a handleMessage() method.  The most universally handled
    message type for actors is the bs.DieMessage.

    # another way to kill the flag from the example above:
    # we can safely call this on any bombsquad type with a 'handleMessage' method.
    # (though its not guaranteed to always have a meaningful effect)
    self.flag.handleMessage(bs.DieMessage())
    """

    # RETAINED actors are kept alive by their owning activity so you don't
    # need to worry about storing references to them
    UNRETAINED = 0
    RETAINED = 1

    def __init__(self,activity,retain):
        """
        Instantiate an actor in the provided activity.

        If 'retain' is bs.Actor.RETAINED, a reference to
        the actor will be stored with the activity so it will
        stay alive as long as the activity exists.

        If 'retain' is bs.Actor.UNRETAINED, the user must retain
        a reference to the actor to keep it alive.
        """
        if activity is None:
            if retain: raise Exception("actor can't be retained without an activity")
            self._activity = None
        else:
            self._activity = weakref.ref(activity)
            if retain: activity._retainActor(self)
            else: activity._addActorWeakRef(self)

    def __del__(self):
        self.handleMessage(DieMessage())

    def handleMessage(self,m):
        """
        General message handling; can be passed any message object.
        """
        pass

    def exists(self):
        """
        Returns True if the actor is still visible or present in some way.
        Note that a dying character should still return True here as long
        as their corpse exists; this is about visiblity, not being alive.

        If this returns False it is assumed the actor can be completely
        deleted without visibly affecting the game; this call is often used
        when 'pruning' lists of actors.

        The default implementation of this method returns 'node.exists()'
        if the actor has a 'node' attr; otherwise it always returns True.
        """

        # as a sensible default, return the existance of self.node
        node = getattr(self,'node',None)
        if node is not None: return node.exists()
        else: return True

    def isAlive(self):
        """
        Returns whether the Actor is 'alive'.
        What this means is up to the actor.
        It is not a requirement for actors to be
        able to die; just that they report whether
        they are alive or not.
        """
        return True

    def getActivity(self):
        """
        Return the bs.Activity this actor is associated with.
        If the activity no longer exists, None is returned.
        """
        try: return self._activity()
        except Exception,e:
            print 'exc in Actor.getActivity():',e
            return None

class NodeActor(Actor):
    """
    category: Game Flow Classes

    A simple bs.Actor which wraps around a single bs.Node and kills
    the node when told to die. This allows you to take advantage of
    standard actor behavior such as dying when no longer referenced,
    so you can do things like kill off a bunch of nodes simply by
    clearing a python list, etc.

    Attributes:
    
       node
          The wrapped node.
    """
    
    #def __init__(self,activity,retain,nodeType,attrs,name=None,delegate=None):
    def __init__(self,activity,retain,node):
        """
        Instantiate with a given bs.Node.
        """
        Actor.__init__(self,activity,retain)
        #self.node = bs.newNode(nodeType,owner=activity,attrs=attrs,name=name,delegate=delegate)
        self.node = node

    def handleMessage(self,m):
        if isinstance(m,DieMessage):
            self.node.delete()




class Session(object):
    """
    category: Game Flow Classes
    
    A Session is the highest level control structure in the game.
    Types of sessions are bs.FreeForAllSession, bs.TeamsSession, and
    bs.CoopSession.

    A Session is responsible for wrangling and transitioning between various
    bs.Activity instances such as mini-games and score-screens, and for
    maintaining state between them (players, teams, score tallies, etc).

    Attributes:

       teams
          All the bs.Teams in the Session. Most things should use the team
          list in bs.Activity; not this.

       players
          All bs.Players in the Session. Most things should use the player
          list in bs.Activity; not this.
    """

    def __init__(self,teamNames=['Good Guys'],
                 teamColors=[(0.6,0.2,1.0)],
                 useTeamColors=True,
                 maxPlayers=8):
        'Instantiate a session with the provided info about teams and max players.'

        import bsLobby
        import bsPowerup
        import bsBomb
        import bsFlag
        import bsSpaz
        import bsScoreSet

        self._useTeams = (teamNames is not None)

        self.teams = []
        self.players = []
        self._nextTeamID = 0
        self._activity = None
        self._ending = False
        self._maxPlayers = maxPlayers
        self._joinSound = bs.getSound("dripity")
        self._playerLeftSound = bs.getSound("playerLeft")

        if self._useTeams:
            for i in range(len(teamColors)):
                team = bs.Team(teamID=self._nextTeamID,
                               name=GameActivity.getTeamNameTranslated(teamNames[i]),
                               color=teamColors[i])
                self.teams.append(team)
                self._nextTeamID += 1
                self.onTeamJoined(team)

        # pull on some stuff to get media loading
        # (this will eventually need to happen per activity and should maybe be optional)
        bs.Spaz.getFactory()
        bs.Flag.getFactory()
        bs.Powerup.getFactory()
        bs.Bomb.getFactory()

        self._lobby = bsLobby.Lobby(teams=self.teams if self._useTeams else None,
                                    useTeamColors=useTeamColors,
                                    delegate=self,vOffset=-150 if isinstance(self,bs.CoopSession) else -30)
        self.scoreSet = bsScoreSet.ScoreSet(self._lobby)

    def onPlayerRequest(self,player):
        """
        Called when a new bs.Player wants to join;
        should return True or False to accept/reject.
        """
        # if we're ending, allow no new players
        if self._ending: return False

        #if len(self.players)+len(self._lobby.getChoosers()) >= self._maxPlayers:
        if len(self.players) >= self._maxPlayers:
            bs.playSound(bs.getSound('error'))
            bs.screenMessage(bs.getResource('playerLimitReachedText').replace('${COUNT}',str(self._maxPlayers)),color=(0.8,0,0))
            return False
        
        bs.playSound(self._joinSound)
        self.players.append(player)
        self._lobby.addChooser(player,activity=self._activity)
        return True


    def onPlayerLeaving(self,player,quiet=False):
        """
        Called when a previously-accepted bs.Player leaves the session.
        """

        # remove them from the game rosters
        if not quiet: bs.playSound(self._playerLeftSound)
            
        if player in self.players:

            # gather the callbacks we'll be making so we can call them
            # *after* we've removed ourself from all lists
            playerLeavingCalls = []
            teamLeavingCalls = []
            
            # this will be None if the player is still in the chooser
            team = player.getTeam()

            # if he was actually in the game, announce his departure
            if not quiet and team is not None:
                bs.screenMessage(bs.getResource('playerLeftText').replace('${PLAYER}',player.getName()))

            # remove him from his team and session lists
            if team is not None:

                # testing.. can remove this eventually
                if isinstance(self,bs.FreeForAllSession):
                    if len(team.players) != 1:
                        bs.printErrorOnce("expected 1 player in FFA team")
                team.players.remove(player)
                
            self.players.remove(player)

            # remove him from the activity list if he's in there
            if player in self._activity.players:
                self._activity.players.remove(player)
                #self._activity.onPlayerLeaving(player)
                playerLeavingCalls.append(self._activity.onPlayerLeaving)

            # if we're a non-team session, remove the team completely
            if not self._useTeams and team is not None:
                self.teams.remove(team)
                #self.onTeamLeaving(team)
                teamLeavingCalls.append(self.onTeamLeaving)
                # same as above; the team may not be in activity..
                if team in self._activity.teams:
                    self._activity.teams.remove(team)
                    #self._activity.onTeamLeaving(team)
                    teamLeavingCalls.append(self._activity.onTeamLeaving)

            # now that both the player and team should be off of all lists, call their leaving methods
            for leavingCall in playerLeavingCalls: leavingCall(player)
            for leavingCall in teamLeavingCalls: leavingCall(team)
        else:
            print 'ERROR: Session.onPlayerLeaving called for player not in our list.'

        # ..and from the chooser
        self._lobby.removeChooser(player)

    def end(self):
        """
        Initiates an end to the session and a return to the main menu.
        Note that this happens asynchronously, allowing the
        session and its activities to shut down gracefully.
        """
        self.setActivity(EndSessionActivity(self))
        self._ending = True # prevents further activity-mucking

    def onTeamJoined(self,team):
        'Called when a new bs.Team joins the session.'
        pass

    def onTeamLeaving(self,team):
        'Called when a bs.Team is leaving the session.'
        pass

    def handleMessage(self,m):
        'General message handling; can be passed any message object.'
        import bsLobby

        if isinstance(m,bsLobby.PlayerReadyMessage):
            self._onPlayerReady(m.chooser)

        elif isinstance(m,EndActivityMessage):
            # if the whole session is shutting down, dont' allow the users to touch anything
            if self._ending: return

            if m.activity is self._activity:

                # if the activity hasn't begun yet, just make note of what to do after we begin
                if not m.activity._hasBegun:
                    if not m.activity._shouldEndImmediately or m.force:
                        m.activity._shouldEndImmediately = True
                        m.activity._shouldEndImmediatelyResults = m.results
                        m.activity._shouldEndImmediatelyDelay = m.delay
                # the activity has begun; get ready to end it..
                else:
                    if (not m.activity._hasEnded) or m.force:
                        m.activity._hasEnded = True
                        self._activityEndTimer = bs.Timer(m.delay,bs.WeakCall(self.onActivityEnd,m.activity,m.results),timeType='real')
                        # also throw a check in there to make sure it's dead within a reasonable amount of time
                        bs.realTimer(m.delay+5000,bs.WeakCall(self._verifyActivityDeath,weakref.ref(m.activity)))

        elif isinstance(m,PlayerProfilesChangedMessage):
            self._lobby.reloadProfiles()
            # for player in self.players:
            #     if player.actor is not None:
            #         updateForPlayer = getattr(player.actor,'updateForPlayer',None)
            #         if updateForPlayer is not None:
            #             updateForPlayer()

        else: Session.handleMessage(self,m)

    def setActivity(self,activity):
        """
        Assign a new current bs.Activity for the session.
        """
        
        if activity is self._activity or self._ending: return

        # this influences whether menus will pause the action or just let it run
        bsInternal._setIsPlaying(activity._isPlayingActivity)

        if activity._isSlowMotion == True:
            bsInternal._setGameSpeed(0.333)
            bsInternal._setSoundPitch(0.4)
        elif activity._isSlowMotion == False:
            bsInternal._setGameSpeed(1.0)
            bsInternal._setSoundPitch(1.0)
        
        activity._hasTransitionedIn = True
        activity.onTransitionIn()

        # we may want to explicitly keep the old activity alive while the new one transitions in
        t = activity._transitionTime
        if t > 0:
            # if we have a previous activity, create a timer that does nothing
            # but holds on to a reference to it until the new one is transitioned in
            def _doNothing(a): pass
            bs.gameTimer(t,bs.Call(_doNothing,self._activity))

        # if we have a current activity, let it transition out before
        # we begin the new one; otherwise begin the new one immediately
        if self._activity is not None:
            self._nextActivity = activity
            self._activity._transitioningOut = True
            self._activity.onTransitionOut()
            # hmm should we not set this until the old one has transitioned out?...
            self._activity = activity
        else:
            self._activity = activity
            activity._begin(self)
            #bs.gameTimer(0,bs.WeakCall(activity._begin,self))

    def getActivity(self):
        'Returns the current activity.'
        return self._activity


    def _verifyActivityDeath(self,activityRef):
        activity = activityRef()
        if activity is not None:
            import types
            import gc
            print 'Activity is not dying when told to:',activity
            print 'Please make sure references get released or use weak references.'
            print 'The following objects are referencing the activity:'
            refs = list(gc.get_referrers(activity))
            i = 1
            def _val(obj):
                s = str(obj)
                if len(s) > 100: s = s[:90]+'...'+s[-10:]
                return s
            for ref in refs:
                if type(ref) is types.FrameType: continue
                print '     ref',i,':',type(ref),':',_val(ref)
                if type(ref) is list or type(ref) is tuple or type(ref) is dict:
                    refs2 = list(gc.get_referrers(ref))
                    j = 1
                    for ref2 in refs2:
                        if type(ref2) is types.FrameType: continue
                        print '        ref '+str(i)+'\'s ref',j,':',type(ref),':',_val(ref2)
                        j += 1
                i += 1

    def _getLobby(self):
        return self._lobby

    def onActivityEnd(self,activity,results):
        """
        Called when the current bs.Activity has ended.
        The session should look at the results and start
        another activity.
        """
        pass
    
    def _runNextActivity(self):
        if self._nextActivity is not None:
            bs.gameTimer(0,bs.WeakCall(self._nextActivity._begin,self))
            self._nextActivity = None

    def _getPlayerIcon(self,player):
        return self._lobby.getPlayerIcon(player)

    def _onPlayerReady(self,chooser):
        'Called when a bs.Player has chosen a character and is ready to join.'

        # during joins, wait till all choosers are ready
        # then create all players at once
        if self._activity._isJoiningActivity:
            if self._lobby.checkAllReady():
                for chooser in self._lobby.getChoosers():
                    player = self._addChosenPlayer(chooser)
                self._lobby.removeAllChoosers()

                # get our next activity going..
                self.onActivityEnd(self._activity,{})
            else:
                return
        # otherwise just add players on the fly
        else:
            player = self._addChosenPlayer(chooser)
            self._lobby.removeChooser(chooser.getPlayer())


    def _addChosenPlayer(self,chooser):

        # pass it to the current activity if it has already begun
        # (otherwise it'll get passed once begin is called)
        passToActivity = (self._activity is not None and self._activity.hasBegun())

        # if we're a non-team game, each player gets their own team
        # (keeps mini-game coding simpler if we can always deal with teams)
        if self._useTeams:
            team = chooser.getTeam()
        else:
            ourTeamID = self._nextTeamID
            team = bs.Team(teamID=ourTeamID,
                           name=chooser.getPlayer().getName(),
                           color=chooser.getColor())
            self.teams.append(team)
            self._nextTeamID += 1
            self.onTeamJoined(team)

            if passToActivity:
                if team in self._activity.teams:
                    print "error: Duplicate team ID in bs.Session._addChosenPlayer"
                self._activity.teams.append(team)
                self._activity.onTeamJoined(team)

        player = chooser.getPlayer() # this is now one and the same

        if not player in self.players:
            bs.printErrorOnce('player not found in session player-list after chooser selection')
            
        player._setData(team=team,
                        character=chooser.getCharacterName(),
                        color=chooser.getColor(),
                        highlight=chooser.getHighlight(),
                        lobby=chooser.getLobby())

        #self.players.append(player)
        self.scoreSet.registerPlayer(player)


        if passToActivity:
            
            # testing.. can remove this eventually
            if isinstance(self,bs.FreeForAllSession):
                if len(player.getTeam().players) != 0:
                    bs.printErrorOnce("expected 0 players in FFA team")
                    
            # dont actually add the player to their team list if we're not in an activity;
            # (players get (re)added to their team lists when the activity begins)
            player.getTeam().players.append(player)
            
            if player in self._activity.players:
                print "error: Duplicate player in bs.Session._addChosenPlayer"
            self._activity.players.append(player)
            self._activity.onPlayerJoined(player)
        return player
        


class EndActivityMessage(object):
    def __init__(self,activity,results=None,delay=0,force=False):
        self.activity = activity
        self.results = results
        self.delay = delay
        self.force = force

class PlayerProfilesChangedMessage(object):
    """Local player profiles may have changed and should be reloaded if they are being used"""
    pass

class Activity(object):
    """
    category: Game Flow Classes

    Units wrangled by a bs.Session.  Examples of Activities include games,
    score-screens, cutscenes, etc. A bs.Session has one 'current' Activity at
    any time, though their existence can overlap during transitions. 

    Attributes:

       settings
          The settings dict passed in when the activity was made.

       teams
          The list of bs.Teams in the activity. This gets populated just before onBegin() is called
          and is updated automatically as players join or leave the game. (at least in free-for-all
          mode where every player gets their own team; in teams mode there are always 2 teams regardless
          of the player count).

       players
          The list of bs.Players in the activity. This gets populated just before onBegin() is called
          and is updated automatically as players join or leave the game.
    """

    def __init__(self,session,settings={}):
        """
        Creates an activity under the provided bs.Session.
        The activity will not be run until bs.Session.setActivity() is called.
        'settings' should be a dict of key/value pairs specific to the activity.
        """
        
        # the activity should preload as much of its media as it can
        # in its constructor, but not actually create any objects yet

        self.settings = settings
        self._session = weakref.ref(session)

        self._hasTransitionedIn = False
        self._hasBegun = False
        self._shouldEndImmediately = False

        # joining activities are for waiting for initial player joins
        # FIXME: should we just test for subclasses of JoiningActivity instead of this?..
        self._isJoiningActivity = False

        # influences the behavior of menus, etc (whether action gets paused)
        self._isPlayingActivity = False
        self._isSlowMotion = False

        # if the activity fades or transitions in, it should set the length of time here
        # so that previous activities will be kept alive for that long (avoiding holes in the screen)
        # also beware that some aspects of the old task won't get cleaned up until the end of this transition
        # so its best to avoid doing any real work until after your transition period is over and
        # your onBegin() is called
        self._transitionTime = 0

        # this gets set once another activity has begun transitioning in but before this one is killed.
        # make sure to not assign player inputs, change music, or anything else disruptive once this gets set.
        self._transitioningOut = False

        # a handy place to put most actors; this list is pruned of dead actors regularly
        # and these actors are insta-killed as the activity is dying.
        self._actorRefs = []
        self._actorWeakRefs = []
        self._ownedNodes = []
        self._ownedNodesCrossFade = []
        self._lastDeadObjectPruneTime = bs.getGameTime()

        # this stuff gets filled in just before onBegin() is called
        self.teams = []
        self.players = []
        self.scoreSet = None

    def _pruneDeadObjects(self):
        self._actorRefs = [a for a in self._actorRefs if a.exists()]
        self._actorWeakRefs = [a for a in self._actorWeakRefs if a() is not None and a().exists()]
        self._ownedNodes = [n for n in self._ownedNodes if n.exists()]
        self._ownedNodesCrossFade = [n for n in self._ownedNodesCrossFade if n.exists()]
        self._lastDeadObjectPruneTime = bs.getGameTime()

    def _retainActor(self,a):
        if not isinstance(a,bs.Actor): raise Exception("non-actor passed to _retainActor")
        if self.hasTransitionedIn() and bs.getGameTime() - self._lastDeadObjectPruneTime > 10000:
            #print 'gt',bs.getGameTime(),'last',self._lastDeadObjectPruneTime
            bs.printErrorOnce('it looks like nodes/actors are not being pruned in your activity;'
                              ' did you call Activity.onTransitionIn() from your subclass?; '+str(self)+' (loc. a)',traceback=False)
        self._actorRefs.append(a)

    def _addActorWeakRef(self,a):
        if not isinstance(a,bs.Actor): raise Exception("non-actor passed to _addActorWeakRef")
        if self.hasTransitionedIn() and bs.getGameTime() - self._lastDeadObjectPruneTime > 10000:
            #print 'gt',bs.getGameTime(),'last',self._lastDeadObjectPruneTime
            bs.printErrorOnce('it looks like nodes/actors are not being pruned in your activity;'
                              ' did you call Activity.onTransitionIn() from your subclass?; '+str(self)+' (loc. b)',traceback=False)
        self._actorWeakRefs.append(weakref.ref(a))

    def _addOwnedNode(self,n,isCrossFade):
        if self.hasTransitionedIn() and bs.getGameTime() - self._lastDeadObjectPruneTime > 10000:
            #print 'gt',bs.getGameTime(),'last',self._lastDeadObjectPruneTime
            bs.printErrorOnce('it looks like nodes/actors are not being pruned in your activity;'
                              ' did you call Activity.onTransitionIn() from your subclass?; '+str(self)+' (loc. c)',traceback=False)
        if isCrossFade:
            self._ownedNodesCrossFade.append(n)
        else:
            self._ownedNodes.append(n)
        
    def __del__(self):

        # do some default cleanup
        try:

            # kick out any players who had a chooser but who handn't joined yet
            if self.getSession() is not None:
                self.getSession()._lobby.removeChooserPlayers()

            # send immediate die messages to our remaining actors
            for actor in self._actorRefs:
                actor.handleMessage(bs.DieMessage(immediate=True))
            for actorRef in self._actorWeakRefs:
                actor = actorRef()
                if actor is not None: actor.handleMessage(bs.DieMessage(immediate=True))

            # reset players
            # (kill any attached actors, clear game-data, etc)
            for player in self.players:
                if player.exists():
                    player._reset()

            # lastly delete remaining nodes
            for node in self._ownedNodes:
                node.delete()

            for team in self.teams:
                team._reset()

        except Exception,e:
            print 'Exception during bs.Activity cleanup:'
            import traceback
            traceback.print_exc(e)

        # tell the session to fire up the next activity
        if self._transitioningOut:
            session = self._session()
            if session: session._runNextActivity()

    def getSession(self):
        """
        Returns the bs.Session this activity belongs to.
        If the session no longer exists, returns None.
        """
        if self._session is None: return None
        return self._session()

    def onPlayerJoined(self,player):
        'Called for all new bs.Players *including* ones in the initial set.'
        pass

    def onPlayerLeaving(self,player):
        'Called just before a departing bs.Player gets pulled from player/team lists.'
        pass

    def onTeamJoined(self,team):
        'Called when a new bs.Team enters the activity (including the initial set).'
        pass

    def onTeamLeaving(self,team):
        'Called when a bs.Team leaves the activity.'
        pass

    def onTransitionIn(self):
        """
        Called when your activity is first becoming visible;
        It should fade in backgrounds, start playing music, etc.
        It does not yet have access to bs.Players or bs.Teams, however,
        until bs.Activity.onBegin() is called; they are still owned
        by the previous activity up until this point.
        """
        self._calledActivityOnTransitionIn = True
        
        # start pruning our transient actors periodically
        self._pruneDeadObjectsTimer = bs.Timer(5000,bs.WeakCall(self._pruneDeadObjects),repeat=True)
        self._pruneDeadObjects()

    def onTransitionOut(self):
        """
        Called when your activity starts transitioning out and a new
        activity is on the way in.  Note that this may happen at any
        time even if finish() has not been called.
        """
        pass


    def onBegin(self):
        """
        Called once the previous bs.Activity has finished transitioning out.
        At this point the activity's initial players and teams are filled in
        and it should begin its actual game logic.
        """
        self._calledActivityOnBegin = True
        
    def handleMessage(self,m):
        'General message handling; can be passed any message object.'
        pass


    def end(self,results=None,delay=0,force=False):
        """
        Commences activity shutdown and delivers results to the bs.Session.
        'delay' is the time delay before the activity actually ends (in milliseconds).
        Further calls to end() will be ignored up until this time, unless
        'force' is True, in which case the new results will replace the old.
        """
        # if we had a standard time-limit that hadnt expired, stop it so it doesnt tick annoyingly
        if hasattr(self,'_standardTimeLimitTime') and self._standardTimeLimitTime > 0:
            self._standardTimeLimitTimer = self._standardTimeLimitText = None

        self.getSession().handleMessage(EndActivityMessage(self,results,delay,force))

        
    def hasTransitionedIn(self):
        'Returns whether onTransitionIn() has been called for this activity.'
        return self._hasTransitionedIn

    def hasBegun(self):
        'Returns whether onBegin() has been called for this activity.'
        return self._hasBegun

    def hasEnded(self):
        'Returns whether end() has been called for this activity.'
        return self._hasEnded

    def isTransitioningOut(self):
        'Returns whether onTransitionOut() has been called for this activity.'
        return self._transitioningOut

    def _begin(self,session):
        'private call to set up onBegin'

        if self._hasBegun:
            bs.printErrorOnce("_begin called twice; this shouldn't happen")
            return
            
        self.scoreSet = session.scoreSet

        # operate on the subset of session players who have passed team/char selection
        players = [p for p in session.players if p.getTeam() is not None]

        # reset all player input
        for p in players: p.resetInput()

        # add teams in one by one and send team-joined messages for each
        for team in session.teams:
            if team in self.teams: raise Exception("Duplicate Team Entry")
            self.teams.append(team)
            self.onTeamJoined(team)

        # now add each player to the activity and to its team's list,
        # and send player-joined messages for each
        for player in players:
            self.players.append(player)

            # testing.. can remove this eventually
            if isinstance(session,bs.FreeForAllSession):
                if len(player.getTeam().players) != 0:
                    bs.printErrorOnce('expected no players in FFA team for player '+
                                      str(player)+'; got '+str([str(p) for p in player.getTeam().players]))
                    
            player.getTeam().players.append(player)
            self.onPlayerJoined(player)

        self._hasBegun = True
        self._hasEnded = False
        self.onBegin()

        
        # make sure that bs.Activity.onTransitionIn() got called at some point
        if not hasattr(self,'_calledActivityOnTransitionIn'):
             bs.printErrorOnce("bs.Activity.onTransitionIn() never got called for "
                               +str(self)
                               +"; did you forget to call it in your onTransitionIn override?")
        else: del self._calledActivityOnTransitionIn
        
        # make sure that bs.Activity.onBegin() got called at some point
        if not hasattr(self,'_calledActivityOnBegin'):
             bs.printErrorOnce("bs.Activity.onBegin() never got called for "
                               +str(self)
                               +"; did you forget to call it in your onBegin override?")
        else: del self._calledActivityOnBegin
            

        # if we've already been told to die, do so now
        if self._shouldEndImmediately: self.end(self._shouldEndImmediatelyResults,self._shouldEndImmediatelyDelay)
        
class EndSessionActivity(Activity):
    """ Special activity to fade out and end the current session """
    def __init__(self,session,settings={}):
        Activity.__init__(self,session,settings)
        self._transitionTime = 250 # keeps prev activity alive while we fade out

    def onTransitionIn(self):
        Activity.onTransitionIn(self)
        bsInternal._fadeLocalScreen(False,time=250)
        bsInternal._lockAllInput()

    def onBegin(self):
        Activity.onBegin(self)

        # ok, for now we want to be all anal-retentive when shutting down sessions
        # and try to ensure that all actors have a chance to clean up their nodes
        # so lets spin here until either that happens or we time out

        # (activities should now remove their own choosers)
        # any choosers we've got up will cause lingering nodes..
        #self.getSession()._lobby.removeChoosers()

        self._checkNum = 0
        self._checkNodesTimer = bs.Timer(100,bs.WeakCall(self._checkNodes),repeat=True)

    def _checkNodes(self):
        nodesRemain = False
        for node in bs.getNodes():
            if not bsUtils._isValidRemainingNode(node):
                nodesRemain = True
                break
        self._checkNum += 1
        # after a second lets print some warnings
        # if nodesRemain: print len(bs.getNodes())
        if self._checkNum == 20:
            bsUtils.printLiveObjectWarnings('2 seconds into EndSessionActivity',ignoreSession=self.getSession(),ignoreActivity=self)
        if (not nodesRemain) or self._checkNum > 40:
            if nodesRemain: print 'Timed out on session shutdown waiting for nodes to die'
            bsInternal._unlockAllInput()
            bsInternal._hostSession('bsMainMenu.MainMenuSession','Main Menu',False)
            

class JoiningActivity(Activity):
    """
    Standard joining activity that shows tips and waits for all players to join.
    """
    def __init__(self,session,settings={}):
        Activity.__init__(self,session,settings)

        # this activity is a special 'joiner' activity
        # - it will get shut down as soon as all players have checked ready
        self._isJoiningActivity = True

    def onTransitionIn(self):
        Activity.onTransitionIn(self)
        self._background = bsUtils.Background(self,bs.Actor.UNRETAINED,
                                              fadeTime=500,startFaded=True,showLogo=True)
        self._tipsText = bsUtils.TipsText(self,bs.Actor.UNRETAINED)
        bs.playMusic('CharSelect')

        self._joinInfo = self.getSession()._getLobby()._createJoinInfo(self)


class TransitionActivity(Activity):
    """
    A simple overlay fade out/in; useful as a bare minimum transition
    between two level based activities.
    """
    def __init__(self,session,settings={}):
        Activity.__init__(self,session,settings)
        self._transitionTime = 500 # keeps prev activity alive while we fade in
        self._isSlowMotion = None # dont change..
    def onTransitionIn(self):
        Activity.onTransitionIn(self)
        self._background = bsUtils.Background(self,bs.Actor.UNRETAINED,
                                              fadeTime=500,startFaded=False,showLogo=False)
    def onBegin(self):
        Activity.onBegin(self)
        # die almost immediately
        bs.gameTimer(100,self.end)

class ScoreScreenActivity(Activity):
    """
    A standard score screen that fades in and shows stuff for a while.
    After a specified delay, player input is assigned to send an EndActivityMessage.
    """

    def __init__(self,session,settings={}):
        Activity.__init__(self,session,settings)
        self._transitionTime = 500
        self._birthTime = bs.getGameTime()
        self._minViewTime = 5000

    def onPlayerJoined(self,player):
        Activity.onPlayerJoined(self,player)
        timeTillAssign = max(0,self._birthTime+self._minViewTime-bs.getGameTime())
        # if we're still kicking at the end of our assign-delay, assign this guy's input to trigger us
        bs.gameTimer(timeTillAssign,bs.WeakCall(self._safeAssign,player))

    def _safeAssign(self,player):
        # just to be extra careful, don't assign if we're transitioning out..
        # (though theoretically that would be ok)
        if not self.isTransitioningOut() and player.exists():
            player.assignInput(('jumpPress','punchPress','bombPress','pickUpPress'),bs.WeakCall(self._playerPress))

    def _playerPress(self):
        if bsInternal._getDemoTimeRemaining() == 0:
            import bsUI
            bsUI.showDemoExpiredUI()
            return
        self.end()

    def onTransitionIn(self,music='Scores',showTips=True):
        Activity.onTransitionIn(self)
        self._background = bsUtils.Background(self,bs.Actor.UNRETAINED,
                                                 fadeTime=500,startFaded=False,showLogo=True)
        if showTips:
            self._tipsText = bsUtils.TipsText(self,bs.Actor.UNRETAINED)
        bs.playMusic(music)

    def onBegin(self,customContinueMessage=None):
        Activity.onBegin(self)
        # pop up a 'press any button to continue' statement after our min-view-time
        # show a 'press any button to continue..' thing after a bit..
        if bs.getEnvironment()['interfaceType'] == 'desktop': s = bs.getResource('pressAnyKeyButtonText')
        else: s = bs.getResource('pressAnyButtonText')

        bsUtils.Text(self,bs.Actor.RETAINED,
                        customContinueMessage if customContinueMessage else s,
                        vAttach='bottom',
                        hAlign='center',
                        flash=True,
                        position=(0,10),
                        scale=0.8,
                        color=(0.5,0.7,0.5,0.5),
                        transition='inBottomSlow',
                        transitionDelay=self._minViewTime)


class GameActivity(Activity):
    """
    category: Game Flow Classes

    Common base class for all game activities; whether of
    the free-for-all, co-op, or teams variety.
    """
    tips = []

    @classmethod
    def createConfigUI(cls,sessionType,config,completionCall):
        """
        Launch a UI to configure settings for this game type under the given bs.Session type.

        'config' should be an existing config dict (specifies 'edit' mode) or None (specifies 'add' mode).

        'completionCall' will be called with a filled-out config dict on success
        or None on cancel.

        Generally subclasses don't need to override this; if they override bs.GameActivity.getSettings()
        and bs.GameActivity.getSupportedMaps() they can just rely on the default implementation here
        which calls those functions.
        """
        import bsUI
        bsUI.standardGameConfigUI(cls,sessionType,config,completionCall)

    @classmethod
    def getName(cls):
        """
        Return a name for this game type in English.
        """
        try: return cls.__module__.replace('_',' ')
        except Exception: return 'Untitled Game'

    @classmethod
    def getNameTranslated(cls):
        """
        Return a translated name for this game type.
        Subclasses should override getName(); not this.
        """
        name = cls.getName()
        try: translated = bs.getResource('translations.gameNames')[name]
        except Exception: translated = None
        if translated is None: translated = name
        return translated

    @classmethod
    def getTeamNameTranslated(cls,name):
        """
        Given a team name, returns a translated version of it.
        """
        try: translated = bs.getResource('translations.teamNames')[name]
        except Exception: translated = None
        if translated is None: translated = name
        return translated

    @classmethod
    def getDescription(cls,sessionType):
        """
        Subclasses should override this to return a description for this activity type
        within the context of the given bs.Session type.
        """
        return ''

    @classmethod
    def getDescriptionTranslated(cls,sessionType):
        """
        Return a translated version of getDescription().
        Sub-classes should override getDescription(); not this.
        """
        description = cls.getDescription(sessionType)
        try: translated = bs.getResource('translations.gameDescriptions')[description]
        except Exception: translated = None
        if translated is None: translated = description
        return translated


    @classmethod
    def getSettings(cls,sessionType):
        """
        Called by the default bs.GameActivity.createConfigUI() implementation;
        should return a dict of config options to be presented
        to the user for the given bs.Session type.

        The format for settings is a list of 2-member tuples consisting
        of a name and a dict of options.

        Available Setting Options:

        'default': This determines the default value as well as the type (int, float, or bool)

        'minValue': Minimum value for int/float settings.
        
        'maxValue': Maximum value for int/float settings.

        'choices': A list of 2-member name/value tuples which the user can toggle through.

        'increment': Value increment for int/float settings.

        # example getSettings() implementation for a capture-the-flag game:
        @classmethod
        def getSettings(cls,sessionType):
            return [("Score to Win",{'default':3,
                                     'minValue':1}),
                    ("Flag Touch Return Time",{'default':0,
                                               'minValue':0,
                                               'increment':1}),
                    ("Flag Idle Return Time",{'default':30,
                                              'minValue':5,
                                              'increment':5}),
                    ("Time Limit",{'default':0,
                                   'choices':[('None',0),
                                              ('1 Minute',60),
                                              ('2 Minutes',120),
                                              ('5 Minutes',300),
                                              ('10 Minutes',600),
                                              ('20 Minutes',1200)]}),
                    ("Respawn Times",{'default':1.0,
                                      'choices':[('Shorter',0.25),
                                                 ('Short',0.5),
                                                 ('Normal',1.0),
                                                 ('Long',2.0),
                                                 ('Longer',4.0)]}),
                    ("Epic Mode",{'default':False})]
        """
        return {}

    @classmethod
    def getSupportedMaps(cls,sessionType):
        """
        Called by the default bs.GameActivity.createConfigUI() implementation;
        should return a list of map names valid for this game-type
        for the given bs.Session type.
        """
        import bsMap
        return bsMap.getMapsSupportingPlayType("melee")

    @classmethod
    def getConfigDescription(cls,config):
        """
        Given a game config dict, return a short description for it
        (translated). This is used when viewing game-lists or
        showing what game is up next in a series.
        """
        import bsMap
        name = cls.getNameTranslated()
        if 'Epic Mode' in config['settings'] and config['settings']['Epic Mode']:
            name = bs.getResource('epicNameFilterText').replace('${NAME}',name)
        s = name+" @ "+bsMap.getTranslatedMapName(bsMap.getFilteredMapName(config['map']))
        return s

    @classmethod
    def supportsSessionType(cls,sessionType):
        """
        Return whether this game type can be played
        in the provided bs.Session subtype.
        """
        return True if issubclass(sessionType,bs.TeamsSession) else False

    def __init__(self,session,mapName,settings={}):
        """
        Instantiates the activity and starts pre-loading the requested map.
        """
        import bsMap
        Activity.__init__(self,session,settings)

        # set some defaults..
        self._isPlayingActivity = True # allow pausing
        self._spawnSound = bs.getSound('spawn')
        self._showKillPoints = True # whether to show points for kills

        # go ahead and get our map loading
        self._mapType = bsMap.getMapClass(mapName)
        self._mapType.preload()
        self._map = None

    def getMap(self):
        """
        Returns the bs.Map in use for this activity.
        """
        if self._map is None: raise Exception("getMap() cannot be called until after onTransitionIn()")
        return self._map
        
    def getInstanceName(self):
        """
        Returns a name for this particular game instance, in English.
        """
        return self.standardInstanceNameFilter(self.getNameTranslated())

    def getInstanceScoreBoardName(self):
        """
        Returns a name for this particular game instance, in English.
        This name is used above the game scoreboard in the corner
        of the screen, so it should be as concise as possible.
        """
        return self.getInstanceName()

    def getInstanceDescription(self):
        """
        Returns a description for this particular game instance, in English. This is
        shown in the center of the screen below the game name at the start of a game.
        It should start with a capital letter and end with a period, and can be a bit
        more verbose than the version returned by getInstanceScoreBoardDescription().

        Note that translation is applied by looking up the specific returned value as a key,
        so the number of returned variations should be limited; ideally just one or two.
        To include arbitrary values in the description, you can return a sequence
        of values in the following form instead of just a string:

        # this will give us something like 'Score 3 goals.' in English
        # and can properly translate to 'Anota 3 goles.' in Spanish.
        # If we just returned the string 'Score 3 Goals' here, there would
        # have to be a translation entry for each specific number. ew.
        return ['Score ${ARG1} goals.', self.settings['Score to Win']]

        This way the first string can be consistently translated, with any arg values
        then substituted into the result. ${ARG1} will be replaced with the first value,
        ${ARG2} with the second, etc.
        """
        return self.getDescription(type(self.getSession()))

    def getInstanceScoreBoardDescription(self):
        """
        Returns a short description for this particular game instance, in English.
        This description is used above the game scoreboard in the corner
        of the screen, so it should be as concise as possible. It should be lowercase
        and should not contain periods or other punctuation.

        Note that translation is applied by looking up the specific returned value as a key,
        so the number of returned variations should be limited; ideally just one or two.
        To include arbitrary values in the description, you can return a sequence
        of values in the following form instead of just a string:

        # this will give us something like 'score 3 goals' in English
        # and can properly translate to 'anota 3 goles' in Spanish.
        # If we just returned the string 'score 3 goals' here, there would
        # have to be a translation entry for each specific number. ew.
        return ['score ${ARG1} goals', self.settings['Score to Win']]

        This way the first string can be consistently translated, with any arg values
        then substituted into the result. ${ARG1} will be replaced with the first value,
        ${ARG2} with the second, etc.

        """
        return ''

    def standardInstanceNameFilter(self,name):
        """
        Apply standard filters to an instance name
        such as changing "GameName" to "Epic GameName" if
        epic-mode is enabled. This is meant to be applied
        post-translation.
        """
        # add bit about epic mode
        return bs.getResource('epicNameFilterText').replace('${NAME}',name) if ('Epic Mode' in self.settings and self.settings['Epic Mode']) else name

    def standardInstanceDescriptionFilter(self,description):
        """
        Apply standard filters to an instance description
        such as changing "Do Stuff to Win." to
        "Do Stuff to Win. In Epic Slow Motion" if
        epic-mode is enabled. This is meant to be applied
        post-translation.
        """
        # add bit about epic mode
        return bs.getResource('epicDescriptionFilterText').replace('${DESCRIPTION}',description) if ('Epic Mode' in self.settings and self.settings['Epic Mode']) else description

    def onTransitionIn(self,music=None):
        """
        Method override; optionally can
        be passed a 'music' string which is the type of
        music to play during the game.
        Note that if the map in use defines music to be
        played, that will be used instead.
        """
        
        Activity.onTransitionIn(self)

        # make our map
        self._map = self._mapType(self,bs.Actor.UNRETAINED)

        # give our map a chance to override the music
        # (for happy-thoughts and other such themed maps)
        overrideMusic = self._mapType.getMusicType()
        if overrideMusic is not None: music = overrideMusic

        if music is not None:  bs.playMusic(music)

    def onBegin(self,nameSound=None):
        """
        Method override; optionally can be passed
        a bs.Sound as 'nameSound' which will be played when
        the game name is shown; otherwise a gong sound will
        be used.
        """
        
        Activity.onBegin(self)

        # we dont do this in onTransitionIn because it may depend on players/teams which arent available until now
        bs.gameTimer(1,bs.WeakCall(self.showScoreBoardInfo))
        bs.gameTimer(1000,bs.WeakCall(self.showInfo,nameSound))
        bs.gameTimer(2500,bs.WeakCall(self._showTip))

        # store some basic info about players present at start time
        self.initialPlayerInfo = [{'name':p.getName(),'character':p.character} for p in self.players]

        # sort this by name so high score lists/etc will be consistent regardless of player join order..
        self.initialPlayerInfo.sort(key=lambda x:x['name'])


    def onPlayerJoined(self,player):
        self.spawnPlayer(player)

    def onPlayerLeaving(self,player):

        # if the player has an actor, lets send it a die-message in a timer; not immediately.
        # that way the player will no longer exist once the message goes through, making
        # it easier for the game to realize the player has left and less likely to make a respawn
        # timer or whatnot.
        actor = player.actor
        if actor is not None:
            # use a strong ref (Call) to make sure the actor doesnt die here due to no refs
            bs.gameTimer(0,bs.Call(actor.handleMessage,bs.DieMessage(how='leftGame')))
        player.setActor(None)
        
        # send a specific death message so its known
        # that they're dying due to leaving the game
        # if player.actor is not None: player.actor.handleMessage(bs.DieMessage(how='leftGame'))
        # player.setActor(None)

    def handleMessage(self,m):
        if isinstance(m,bs.PlayerSpazDeathMessage):
            # inform our score-set of the demise
            player = m.spaz.getPlayer()

            # temp - remove this at some point
            if player is None:
                bs.printErrorOnce('FIXME: getPlayer() should no longer ever be returning None')
                return
                
            if not player.exists(): return # this death may be due to the player leaving

            killer = m.killerPlayer

            self.scoreSet.playerLostSpaz(player,killed=m.killed,killer=killer)

            # award the killer points if he's on a different team
            if killer is not None and killer.getTeam() is not player.getTeam():
                pts,importance = m.spaz.getDeathPoints(m.how)
                if not self.hasEnded():
                    self.scoreSet.playerScored(killer,pts,kill=True,victimPlayer=player,importance=importance,showPoints=self._showKillPoints)


    def showScoreBoardInfo(self):
        """
        Creates the game info display
        in the top left corner showing the name
        and short description of the game.
        """
        
        sbName = self.getInstanceScoreBoardName()

        # the description can be either a string or a sequence with args to swap in post-translation
        sbDesc = self.getInstanceScoreBoardDescription()
        if type(sbDesc) in [unicode,str]: sbDesc = [sbDesc] # handle simple string case
        if type(sbDesc[0]) not in [unicode,str]: raise Exception("Invalid format for instance description")
        try: translation = bs.getResource('translations.gameDescriptions')[sbDesc[0]]
        except Exception: translation = None
        if translation is None: translation = sbDesc[0]
        for i in range(len(sbDesc)-1):
            translation = translation.replace('${ARG'+str(i+1)+'}',str(sbDesc[i+1]))
        sbDesc = translation

        y = -34 if sbDesc == '' else -20
        y -= 16
        self._gameScoreBoardNameText = bs.NodeActor(self,bs.Actor.UNRETAINED,
                                                    bs.newNode("text",
                                                               name='Game Scoreboard Name',
                                                               owner=self,
                                                               attrs={'text':sbName,
                                                                      'maxWidth':300,
                                                                      'translate':(15,y) if isinstance(self.getSession(),bs.FreeForAllSession) else (15,y),
                                                                      'hAttach':"left",
                                                                      'vAttach':"top",
                                                                      'vAlign':'bottom',
                                                                      'color':(1.0,1.0,1.0,1.0),
                                                                      'shadow':0.6,
                                                                      'scale':[1.1]}))
        
        bsUtils.animate(self._gameScoreBoardNameText.node,'opacity',{0:0.0,1000:1.0})

        self._gameScoreBoardDescriptionText = bs.NodeActor(self,bs.Actor.UNRETAINED,
                                                           bs.newNode("text",
                                                                      name='Game Scoreboard Desc',
                                                                      owner=self,
                                                                      attrs={'text':sbDesc,
                                                                             'maxWidth':480,
                                                                             'translate':(17,-44+10) if isinstance(self.getSession(),bs.FreeForAllSession) else (17,-44+10),
                                                                             'scale':[0.7],
                                                                             'hAttach':"left",
                                                                             'vAttach':"top",
                                                                             'vAlign':'top',
                                                                             'shadow':0.7,
                                                                             'flatness':0.8,
                                                                             'color':(0.9,0.9,0.9,1.0)}))
        
        bsUtils.animate(self._gameScoreBoardDescriptionText.node,'opacity',{0:0.0,1000:1.0})

    def showInfo(self,nameSound):
        """ show the game description """
        name = self.getInstanceName()
        bsUtils.ZoomText(self,bs.Actor.RETAINED,
                            name,maxWidth=800,lifespan=2500,jitter=2.0,position=(0,180),
                            flash=False,color=(0.93*1.25,0.9*1.25,1.0*1.25),trailColor=(0.15,0.05,1.0,0.0))
        bs.gameTimer(200,bs.Call(bs.playSound,nameSound if nameSound is not None else bs.getSound('gong')))
            
        # the description can be either a string or a sequence with args to swap in post-translation
        desc = self.getInstanceDescription()
        if type(desc) in [unicode,str]: desc = [desc] # handle simple string case
        if type(desc[0]) not in [unicode,str]: raise Exception("Invalid format for instance description")
        try: translation = bs.getResource('translations.gameDescriptions')[desc[0]]
        except Exception: translation = None
        if translation is None: translation = desc[0]
        for i in range(len(desc)-1):
            translation = translation.replace('${ARG'+str(i+1)+'}',str(desc[i+1]))

        # filter for epic-mode and this and that
        translation = self.standardInstanceDescriptionFilter(translation)

        d = bs.newNode('text',
                       owner=self,
                       attrs={'vAttach':'center',
                              'hAttach':'center',
                              'hAlign':'center',
                              'color':(1,1,1,1),
                              'shadow':0.5,
                              'translate':(0,80),
                              'scale':[1.2],
                              'maxWidth':700,
                              'text':translation},name='Game Description')
        c = bs.newNode("combine",owner=d,attrs={'input0':1.0,'input1':1.0,'input2':1.0})
        c.connectAttr('output',d,'color')
        keys = {500:0,1000:1.0,2500:1.0,4000:0.0}
        bsUtils.animate(c,"input3",keys)
        bs.gameTimer(4000,d.delete)


    def _showTip(self):

        # if theres any tips left on the global list, display one..
        if len(self.tips) > 0:
            tip = self.tips.pop(random.randrange(len(self.tips)))
            tipTitle = bs.getResource('tipText')+':'
            icon = None
            sound = None
            if type(tip) == dict:
                if 'icon' in tip: icon = tip['icon']
                if 'sound' in tip: sound = tip['sound']
                tip = tip['tip']
            tip = bs.translate('tips',tip)

            # a few subs..
            #tip = tip.replace('${PICKUP}','\xc2\x85')
            #tip = tip.replace('${PICKUP}','\xee\x80\x86')
            tip = tip.replace('${PICKUP}',bs.getSpecialChar('topButton'))
            baseTranslate = (75,50)
            tipScale = 0.8
            tipTitleScale = 1.2
            tipWidthRaw = bs.getStringWidth(tip)
            if tipWidthRaw > 950: tipScale *= (950.0/tipWidthRaw)
            tipWidth = tipWidthRaw*tipScale

            t = bs.newNode('text',
                           owner=self,
                           attrs={'text':tip,
                                  'scale':[tipScale],
                                  'translate':(baseTranslate[0]-tipWidth*0.5,baseTranslate[1]),
                                  'hAlign':'left',
                                  'shadow':0.5,
                                  'flatness':0.5,
                                  'vAlign':'center',
                                  'vAttach':'bottom'},
                           name='Tip')
            t2 = bs.newNode('text',
                            owner=t,
                            attrs={'text':tipTitle,
                                   'scale':[tipTitleScale],
                                   'translate':(baseTranslate[0]-tipWidth*0.5-(20 if icon is None else 82),baseTranslate[1]+2),
                                   'hAlign':'right',
                                   'shadow':0.5,
                                   'flatness':0.5,
                                   'maxWidth':140,
                                   'vAlign':'center',
                                   'vAttach':'bottom'},
                            name='Tip2')
            if icon is not None:
                img = bs.newNode('image',
                                 owner=self,
                                 attrs={'texture':icon,
                                        'translate':(baseTranslate[0]-tipWidth*0.5-40,baseTranslate[1]+1),
                                        'scale':(50,50),
                                        'opacity':1.0,
                                        'color':(1,1,1),
                                        'absoluteScale':True,
                                        'attach':'bottomCenter'})
                bsUtils.animate(img,'opacity',{0:0,1000:1,4000:1,5000:0})
                bs.gameTimer(5000,img.delete)
            if sound is not None:
                bs.playSound(sound)

            c = bs.newNode("combine",owner=t,attrs={'input0':1.0,'input1':0.8,'input2':1.0})
            c.connectAttr('output',t,'color')
            c.connectAttr('output',t2,'color')
            bsUtils.animate(c,'input3',{0:0,1000:1,4000:1,5000:0})
            bs.gameTimer(5000,t.delete)

    def endGame(self):
        """
        Tells the game to wrap itself up and call bs.Activity.end() immediately.
        This method should be overridden by subclasses.

        A game should always be prepared to end and deliver results, even if
        there is no 'winner' yet; this way things like the standard time-limit
        (bs.GameActivity.setupStandardTimeLimit()) will work with the game.
        """
        pass

    def spawnPlayerIfExists(self,player):
        """
        A utility method which calls self.spawnPlayer() *only* if the bs.Player
        provided still exists; handy for use in timers and whatnot.

        There is no need to override this; just override spawnPlayer().
        """
        if player.exists(): self.spawnPlayer(player)

    def spawnPlayer(self,player):
        """
        Spawn *something* for the provided bs.Player.
        The default implementation simply calls spawnPlayerSpaz().
        """

        if not player.exists():
            bs.printErrorOnce('spawnPlayer() called for nonexistant player')
            return
        
        return self.spawnPlayerSpaz(player)

    def respawnPlayer(self,player,respawnTime=None):
        """
        Given a bs.Player, sets up a standard respawn timer,
        along with the standard counter display, etc.
        At the end of the respawn period spawnPlayer() will
        be called if the Player still exists.
        An explicit 'respawnTime' can optionally be provided
        (in milliseconds).
        """
        
        if player is None or not player.exists():
            if player is None: bs.printErrorOnce('None passed as player to respawnPlayer()')
            else: bs.printErrorOnce('Nonexistant bs.Player passed to respawnPlayer(); call player.exists() to make sure a player is still there.')
            return

        if respawnTime is None:
            if len(player.getTeam().players) == 1: respawnTime = 3000
            elif len(player.getTeam().players) == 2: respawnTime = 5000
            elif len(player.getTeam().players) == 3: respawnTime = 6000
            else: respawnTime = 7000

        # if this standard setting is present, factor it in
        if 'Respawn Times' in self.settings: respawnTime *= self.settings['Respawn Times']

        respawnTime = int(max(1000,respawnTime))
        if respawnTime%1000 != 0: respawnTime -= respawnTime%1000 # we want whole seconds

        if player.actor and not self.hasEnded():
            import bsSpaz
            player.gameData['respawnTimer'] = bs.Timer(respawnTime,bs.WeakCall(self.spawnPlayerIfExists,player))
            player.gameData['respawnIcon'] = bsSpaz.RespawnIcon(player,self,respawnTime)
        

    def spawnPlayerSpaz(self,player,position=(0,0,0),angle=None):
        """
        Create and wire up a bs.PlayerSpaz for the provide bs.Player.
        """
        name = player.getName()
        color = player.color
        highlight = player.highlight

        lightColor = bsUtils.getNormalizedColor(color)
        displayColor = bs.getSafeColor(color)
        spaz = bs.PlayerSpaz(self,bs.Actor.UNRETAINED,
                                 color=color,
                                 highlight=highlight,
                                 character=player.character,
                                 player=player,
                                 powerupsExpire=isinstance(self.getSession(),bs.CoopSession))
        player.setActor(spaz)

        # we want a bigger area-of-interest in co-op mode
        # if isinstance(self.getSession(),bs.CoopSession): spaz.node.areaOfInterestRadius = 5.0
        # else: spaz.node.areaOfInterestRadius = 5.0
        
        spaz.node.name = name
        spaz.node.nameColor = displayColor
        spaz.connectControlsToPlayer()
        self.scoreSet.playerGotNewSpaz(player,spaz)
        spaz.handleMessage(bs.StandMessage(position,angle if angle is not None else random.uniform(0,360)))
        t = bs.getGameTime()
        bs.playSound(self._spawnSound,1,position=(spaz.node,"translate"))
        light = bs.newNode("light",owner=self,attrs={"color":lightColor})
        spaz.node.connectAttr("translate",light,"translate")
        bsUtils.animate(light,"intensity",{0:0,250:1,500:0})
        bs.gameTimer(500,light.delete)
        return spaz

    def projectFlagStand(self,pos):
        """
        Projects a flag-stand onto the ground at the given position.
        Useful for games such as capture-the-flag to show where a
        movable flag originated from.
        """
        # need to do this in a timer for it to work.. need to look into that.
        bs.gameTimer(1,bs.WeakCall(self._projectFlagStand,pos[:3]))
        
    def _projectFlagStand(self,pos):
        bs.emitBGDynamics(position=pos,emitType='flagStand')

    def _standardTimeLimitGetStr(self):
        value = self._standardTimeLimitTime
        return bsUtils.getTimeString(value*1000,centi=False)

    def setupStandardPowerupDrops(self,enableTNT=True):
        """
        Create standard powerup drops for the current map.
        """
        import bsPowerup
        self._powerupDropTimer = bs.Timer(bsPowerup.defaultPowerupInterval,bs.WeakCall(self._standardDropPowerups),repeat=True)
        self._standardDropPowerups()
        if enableTNT:
            self._tntObjs = {}
            self._tntDropTimer = bs.Timer(5500,bs.WeakCall(self._standardDropTnt),repeat=True)
            self._standardDropTnt()


    def _standardDropPowerup(self,index,expire=True):
        import bsPowerup
        bsPowerup.Powerup(self,bs.Actor.RETAINED,translate=self.getMap().powerupSpawnPoints[index],
                          powerupType=bs.Powerup.getFactory().getRandomPowerupType(),expire=expire)

    def _standardDropPowerups(self):
        """
        Standard powerup drop.
        """
        # drop one powerup per point
        pts = self.getMap().powerupSpawnPoints
        for i,pt in enumerate(pts):
            bs.gameTimer(i*400,bs.WeakCall(self._standardDropPowerup,i))

    def _standardDropTnt(self):
        """
        Standard tnt drop.
        """
        # drop tnt on the map for any tnt location with no existing tnt box
        for i,pt in enumerate(self.getMap().tntPoints):
            if not i in self._tntObjs: self._tntObjs[i] = {'absentTicks':9999,'obj':None}
            tntObj = self._tntObjs[i]

            # respawn once its been dead for a while..
            if tntObj['obj'] is None or not tntObj['obj'].exists():
                tntObj['absentTicks'] += 1
                if tntObj['absentTicks'] > 3:
                    tntObj['obj'] = bs.Bomb(self,bs.Actor.UNRETAINED,
                                            translate=pt,bombType='tnt')
                    tntObj['absentTicks'] = 0

    def setupStandardTimeLimit(self,duration):
        """
        Create a standard game time-limit given the provided
        duration in seconds.
        This will be displayed at the top of the screen.
        If the time-limit expires, endGame() will be called.
        """

        if duration <= 0: return
        self._standardTimeLimitTime = int(duration)
        self._standardTimeLimitTimer = bs.Timer(1000,bs.WeakCall(self._standardTimeLimitTick),repeat=True)
        self._standardTimeLimitText = bs.NodeActor(self,bs.Actor.UNRETAINED,
                                                   bs.newNode('text',
                                                              owner=self,
                                                              attrs={'vAttach':'top',
                                                                     'hAttach':'center',
                                                                     'hAlign':'left',
                                                                     'color':(1.0,1.0,1.0,0.5),
                                                                     'translate':(-25,-30),
                                                                     'flatness':1.0,
                                                                     #'shadow':1.0,
                                                                     'scale':[0.9],
                                                                     'text':self._standardTimeLimitGetStr()}))
    def _standardTimeLimitTick(self):
        self._standardTimeLimitTime -= 1
        if self._standardTimeLimitTime <= 10:
            if self._standardTimeLimitTime == 10:
                self._standardTimeLimitText.node.scale = [1.3]
                self._standardTimeLimitText.node.translate = (-30,-45)
                c = bs.newNode('combine',owner=self._standardTimeLimitText.node)
                c.connectAttr('output',self._standardTimeLimitText.node,'color')
                bsUtils.animate(c,"input0",{0:1, 150:1},loop=True)
                bsUtils.animate(c,"input1",{0:1, 150:0.5},loop=True)
                bsUtils.animate(c,"input2",{0:0.1, 150:0.0},loop=True)
                c.input3 = 1.0
            bs.playSound(bs.getSound('tick'))
        if self._standardTimeLimitTime <= 0:
            self._standardTimeLimitTimer = None
            self.endGame()
            n = bs.newNode('text',
                           owner=self,
                           attrs={'vAttach':'top','hAttach':'center',
                                  'hAlign':'center','color':(1,0.7,0,1),
                                  'translate':(0,-90),'scale':[1.2],'text':bs.getResource('timeExpiredText')})
            bs.playSound(bs.getSound('refWhistle'))
            bsUtils.animate(n,"scale",{0:0.0, 100:1.4, 150:1.2})

        self._standardTimeLimitText.node.text = self._standardTimeLimitGetStr()

    def showZoomMessage(self,message,color=(0.9,0.4,0.0),scale=0.8,duration=2000,trail=False):
        """
        Show the standard zooming text used to announce
        game names and winners.
        """
        # reserve a spot on the screen (in case we get multiple of these so they dont overlap)
        try: times = self._zoomMessageTimes
        except Exception: self._zoomMessageTimes = {}
        i = 0
        curTime = bs.getGameTime()
        while True:
            if i not in self._zoomMessageTimes or self._zoomMessageTimes[i] < curTime:
                self._zoomMessageTimes[i] = curTime + duration
                break
            i += 1
        bsUtils.ZoomText(self,bs.Actor.RETAINED,
                            message,lifespan=duration,jitter=2.0,position=(0,200-i*100),scale=scale,maxWidth=800,
                            trail=trail,color=color)

        
    def cameraFlash(self,duration=999):
        """
        Create a strobing camera flash effect
        as seen when a team wins a game.
        """
        xSpread = 10
        ySpread = 5
        positions = [[-xSpread,-ySpread],[0,-ySpread],[0,ySpread],[xSpread,-ySpread],[xSpread,ySpread],[-xSpread,ySpread]]
        times = [0,2700,1000,1800,500,1400]

        # store this on our activity so we only have one at a time
        self._cameraFlash = []
        for i in range(6):
            light = bs.NodeActor(self,bs.Actor.UNRETAINED,
                                 bs.newNode("light",
                                            owner=self,
                                            attrs={'translate':(positions[i][0],0,positions[i][1]),
                                                   'radius':1.0,
                                                   'lightVolumes':False,
                                                   'heightAttenuated':False,
                                                   'color':(0.2,0.2,0.8)}))
            s = 1.87
            iScale = 1.3
            tcombine = bs.newNode("combine",owner=light.node,attrs={'input0':positions[i][0],'input1':0,'input2':positions[i][1]})
            tcombine.connectAttr('output',light.node,'translate')
            x = positions[i][0]
            y = positions[i][1]
            spd = 0.5 + random.random()
            spd2 = 0.5 + random.random()
            bsUtils.animate(tcombine,'input0',{0:x+0, 69*spd:x+10.0, 143*spd:x-10.0, 201*spd:x+0},loop=True)
            bsUtils.animate(tcombine,'input2',{0:y+0, 150*spd2:y+10.0, 287*spd2:y-10.0, 398*spd2:y+0},loop=True)
            
            bsUtils.animate(light.node,"intensity",{0:0, 20*s:0, 50*s:0.8*iScale, 80*s:0, 100*s:0},loop=True,offset=times[i])
            bs.gameTimer(int(times[i]+random.randint(1,duration)*40*s),light.node.delete)
            self._cameraFlash.append(light)

